"""Tests for CoBib's commands."""
# pylint: disable=unused-argument, redefined-outer-name

import json
import os
import re
import subprocess
import sys
from datetime import datetime
from io import StringIO
from itertools import zip_longest
from pathlib import Path
from shutil import rmtree

import pytest
from cobib import commands
from cobib.config import CONFIG
from cobib.database import read_database


def assert_git_commit_message(command, args):
    """Assert the last auto-generated git commit message."""
    # get last commit message
    proc = subprocess.Popen(['git', '-C', '/tmp/cobib_test', 'show',
                             '--format=format:%B', '--no-patch', 'HEAD'],
                            stdout=subprocess.PIPE)
    message, _ = proc.communicate()
    # decode it
    message = message.decode('utf-8').split('\n')
    # assert subject line
    assert f'Auto-commit: {command.title()}Command' in message[0]
    # assert args
    args = json.dumps(args, indent=2, default=str)
    for ref, truth in zip(args.split('\n'), message[2:]):
        assert ref == truth


@pytest.fixture
def setup():
    """Setup."""
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    CONFIG.validate()
    read_database()


def test_set_config(setup):
    """Test config setting.

    Args:
        setup: runs pytest fixture.
    """
    # from setup
    assert CONFIG.config['DATABASE']['file'] == './test/example_literature.yaml'
    # change back to default
    CONFIG.set_config()
    CONFIG.validate()
    assert CONFIG.config['DATABASE']['file'] == \
        os.path.expanduser('~/.local/share/cobib/literature.yaml')


@pytest.fixture(params=[False, True])
def init_setup(request):
    """Setup for InitCommand testing."""
    # use temporary config
    tmp_config = "[DATABASE]\nfile=/tmp/cobib_test/database.yaml\n"
    if request.param:
        tmp_config += 'git=True\n'
    with open('/tmp/cobib_test_config.ini', 'w') as file:
        file.write(tmp_config)
    # ensure configuration is empty
    CONFIG.config = {}
    # load config
    CONFIG.set_config(Path('/tmp/cobib_test_config.ini'))
    CONFIG.validate()
    # yielding the arguments allows re-using them inside of the actual test function
    yield request.param
    # clean up file system
    os.remove('/tmp/cobib_test_config.ini')
    if request.param:
        rmtree('/tmp/cobib_test/.git')


@pytest.mark.parametrize(['safe'], [
        [False],
        [True],
    ])
def test_init(init_setup, safe):
    """Test init command."""
    git = init_setup
    if safe:
        # fill database file
        with open('/tmp/cobib_test/database.yaml', 'w') as file:
            file.write('test')
    # store current time
    now = float(datetime.now().timestamp())
    # try running init
    commands.InitCommand().execute(['--git'] if git else [])
    if safe:
        # check database file still contains 'test'
        with open('/tmp/cobib_test/database.yaml', 'r') as file:
            assert file.read() == 'test'
    else:
        # check creation time of temporary database file
        ctime = os.stat('/tmp/cobib_test/database.yaml').st_ctime
        # assert these times are close
        assert ctime - now < 0.1 or now - ctime < 0.1
    if git:
        # check creation time of temporary database git folder
        ctime = os.stat('/tmp/cobib_test/.git').st_ctime
        # assert these times are close
        assert ctime - now < 0.1 or now - ctime < 0.1
        # and assert that it is indeed a folder
        assert os.path.isdir('/tmp/cobib_test/.git')
        # assert the git commit message
        assert_git_commit_message('init', {'git': True})
    # clean up file system
    os.remove('/tmp/cobib_test/database.yaml')


@pytest.mark.parametrize(['args', 'expected'], [
        [[], ['einstein', 'latexcompanion', 'knuthwebsite']],
        [['-r'], ['knuthwebsite', 'latexcompanion', 'einstein']],
        [['-s', 'year'], ['einstein', 'knuthwebsite', 'latexcompanion']],
        [['-r', '-s', 'year'], ['latexcompanion', 'knuthwebsite', 'einstein']],
    ])
def test_list(setup, args, expected):
    """Test list command.

    Args:
        setup: runs pytest fixture.
        args: arguments for the list command call.
        expected: expected result.
    """
    # redirect output of list to string
    file = StringIO()
    tags = commands.ListCommand().execute(args, out=file)
    assert tags == expected
    for line in file.getvalue().split('\n'):
        if line.startswith('ID') or all([c in '- ' for c in line]):
            # skip table header
            continue
        assert line.split()[0] in expected


def test_list_with_missing_keys(setup):
    """Asserts issue #1 is fixed.

    When a key is queried which is not present in all entries, the list command should return
    normally.

    Args:
        setup: runs pytest fixture.
    """
    # redirect output of list to string
    file = StringIO()
    tags = commands.ListCommand().execute(['++year', '1905'], out=file)
    expected = ['einstein']
    assert tags == expected
    for line in file.getvalue().split('\n'):
        if line.startswith('ID') or all([c in '- ' for c in line]):
            # skip table header
            continue
        assert line.split()[0] in expected


def test_show(setup):
    """Test show command.

    Args:
        setup: runs pytest fixture.
    """
    file = StringIO()
    commands.ShowCommand().execute(['einstein'], out=file)
    with open('./test/example_literature.bib', 'r') as expected:
        for line, truth in zip_longest(file.getvalue().split('\n'), expected):
            if not line:
                continue
            assert line == truth.strip('\n')


@pytest.fixture
def open_setup():
    """Setup for OpenCommand testing."""
    # ensure configuration is empty
    CONFIG.config = {}
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    CONFIG.validate()
    # NOTE: normally you would never trigger an Add command before reading the database but in this
    # controlled testing scenario we can be certain that this is fine
    commands.AddCommand().execute(['-b', './test/dummy_multi_file_entry.bib'])
    read_database()
    yield setup
    # clean up
    commands.DeleteCommand().execute(['dummy_multi_file_entry'])


def test_open(open_setup):
    """Test open command.

    Args:
        open_setup: runs pytest fixture.
    """
    # pylint: disable=missing-class-docstring
    class DummyStdin:
        # pylint: disable=missing-function-docstring
        def readline(self):
            # pylint: disable=no-self-use
            return '\n'
    # replace sys.stdout and sys.stdin
    original_stdout = sys.stdout
    original_stdin = sys.stdin
    sys.stdout = StringIO()
    sys.stdin = DummyStdin()
    commands.OpenCommand().execute(['dummy_multi_file_entry'])
    expected = [
        "  1: [file] /tmp/a.txt",
        "  2: [file] /tmp/b.txt",
        "  3: [url] https://www.duckduckgo.com",
        "  4: [url] https://www.google.com",
        "Entry to open [Type 'help' for more info]: ",
    ]
    for line, truth in zip_longest(sys.stdout.getvalue().split('\n'), expected):
        assert line == truth
    # clean up
    sys.stdout = original_stdout
    sys.stdin = original_stdin


@pytest.fixture
def database_setup(init_setup):
    """Initialize a database on top of the init_setup fixture."""
    git = init_setup
    # initialize database
    # NOTE: if the InitCommand fails, all tests depending on this will fail, too
    commands.InitCommand().execute(['--git'] if git else [])
    # freshly read in database to overwrite anything that was read in during setup()
    read_database(fresh=True)
    # yield the parameter to allow re-use in actual test function
    yield git
    # clean up file system
    os.remove('/tmp/cobib_test/database.yaml')


def test_add(database_setup):
    """Test add command."""
    git = database_setup
    # add some data
    commands.AddCommand().execute(['-b', './test/example_literature.bib'])
    # compare with reference file
    with open('/tmp/cobib_test/database.yaml', 'r') as file:
        with open('./test/example_literature.yaml', 'r') as expected:
            for line, truth in zip_longest(file, expected):
                assert line == truth
    if git:
        # assert the git commit message
        with open('./test/example_literature.bib', 'r') as bibtex:
            assert_git_commit_message('add', {
                'label': None,
                'file': None,
                'arxiv': None,
                'bibtex': bibtex,
                'doi': None,
                'isbn': None,
                'tags': [],
            })


def test_add_overwrite_label(database_setup):
    """Test add command while specifying a label manually.

    Regression test against #4.
    """
    # add some data
    commands.AddCommand().execute(['-b', './test/example_literature.bib'])
    # add potentially duplicate entry
    commands.AddCommand().execute(['-b', './test/example_duplicate_entry.bib',
                                   '--label', 'duplicate_resolver'])
    # compare with reference file
    with open('./test/example_literature.yaml', 'r') as expected:
        true_lines = expected.readlines()
    with open('./test/example_duplicate_entry.yaml', 'r') as extra:
        true_lines += extra.readlines()
    with open('/tmp/cobib_test/database.yaml', 'r') as file:
        for line, truth in zip_longest(file, true_lines):
            assert line == truth


@pytest.mark.parametrize(['labels'], [
        [['knuthwebsite']],
        [['knuthwebsite', 'latexcompanion']],
    ])
def test_delete(database_setup, labels):
    """Test delete command."""
    git = database_setup
    # NOTE: DeleteCommand depends on AddCommand to work. While this is not so nice for the
    # unittests, it is the easiest method of testing all git and non-git scenarios.
    commands.AddCommand().execute(['-b', './test/example_literature.bib'])
    # delete some data
    # NOTE: for testing simplicity we delete the last entry
    commands.DeleteCommand().execute(labels)
    with open('/tmp/cobib_test/database.yaml', 'r') as file:
        with open('./test/example_literature.yaml', 'r') as expected:
            # NOTE: do NOT use zip_longest to omit last entry (thus, we deleted the last one)
            for line, truth in zip(file, expected):
                assert line == truth
            with pytest.raises(StopIteration):
                file.__next__()
    if git:
        # assert the git commit message
        assert_git_commit_message('delete', {'labels': labels})


@pytest.mark.parametrize(['modification', 'filters', 'selection'], [
        ['tags:test', ['einstein'], True],
        ['tags:test', ['++ID', 'einstein'], False],
    ])
def test_modify(database_setup, modification, filters, selection, append=False):
    """Test modify command."""
    git = database_setup
    # NOTE: again, we depend on AddCommand to work.
    commands.AddCommand().execute(['-b', './test/example_literature.bib'])
    read_database(fresh=True)
    # modify some data
    args = [modification, '--'] + filters
    if selection:
        args = ['-s'] + args
    if append:
        args = ['-a'] + args
    commands.ModifyCommand().execute(args)
    assert CONFIG.config['BIB_DATA']['einstein'].data['tags'] == 'test'
    if git:
        # assert the git commit message
        assert_git_commit_message('modify', {'append': append,
                                             'selection': selection,
                                             'modification': modification.split(':'),
                                             'filter': filters})


# TODO: figure out some very crude and basic way of testing this
def test_edit():
    """Test edit command."""
    pytest.skip("There is currently no meaningful way of testing this.")


def test_undo(database_setup, caplog):
    """Test undo command."""
    git = database_setup
    if not git:
        commands.UndoCommand().execute([])
        for record in caplog.records:
            if record.name == 'cobib.commands.undo':
                assert record.levelname == 'ERROR'
                assert 'git-tracking' in record.msg
    else:
        commands.AddCommand().execute(['-b', './test/example_literature.bib'])
        commands.UndoCommand().execute([])
        assert os.stat('/tmp/cobib_test/database.yaml').st_size == 0
        # get last commit message
        proc = subprocess.Popen(['git', '-C', '/tmp/cobib_test', 'show',
                                 '--format=format:%B', '--no-patch', 'HEAD'],
                                stdout=subprocess.PIPE)
        message, _ = proc.communicate()
        # decode it
        message = message.decode('utf-8').split('\n')
        # assert subject line
        assert 'Undo' in message[0]


def test_redo(database_setup, caplog):
    """Test redo command."""
    git = database_setup
    if not git:
        commands.RedoCommand().execute([])
        for record in caplog.records:
            if record.name == 'cobib.commands.redo':
                assert record.levelname == 'ERROR'
                assert 'git-tracking' in record.msg
    else:
        commands.AddCommand().execute(['-b', './test/example_literature.bib'])
        commands.UndoCommand().execute([])
        commands.RedoCommand().execute([])
        with open('/tmp/cobib_test/database.yaml', 'r') as file:
            with open('./test/example_literature.yaml', 'r') as expected:
                for line, truth in zip_longest(file, expected):
                    assert line == truth
        # get last commit message
        proc = subprocess.Popen(['git', '-C', '/tmp/cobib_test', 'show',
                                 '--format=format:%B', '--no-patch', 'HEAD'],
                                stdout=subprocess.PIPE)
        message, _ = proc.communicate()
        # decode it
        message = message.decode('utf-8').split('\n')
        # assert subject line
        assert 'Redo' in message[0]


def test_export(setup):
    """Test export command.

    Args:
        setup: runs pytest fixture.
    """
    commands.ExportCommand().execute(['-b', '/tmp/cobib_test_export.bib'])
    with open('/tmp/cobib_test_export.bib', 'r') as file:
        with open('./test/example_literature.bib', 'r') as expected:
            for line, truth in zip_longest(file, expected):
                if truth[0] == '%':
                    # ignore comments
                    continue
                assert line == truth
    # clean up file system
    os.remove('/tmp/cobib_test_export.bib')


def test_export_selection(setup):
    """Test the `selection` interface of the export command.

    Args:
        setup: runs pytest fixture.
    """
    commands.ExportCommand().execute(['-b', '/tmp/cobib_test_export_s.bib', '-s', '--', 'einstein'])
    with open('/tmp/cobib_test_export_s.bib', 'r') as file:
        with open('./test/example_literature.bib', 'r') as expected:
            for line, truth in zip_longest(file, expected):
                print(line, truth)
                if truth[0] == '%':
                    # ignore comments
                    continue
                if truth.strip() == '@book{latexcompanion,':
                    # reached next entry
                    break
                assert line == truth
    # clean up file system
    os.remove('/tmp/cobib_test_export_s.bib')


@pytest.mark.parametrize(['args', 'expected', 'config_overwrite'], [
        [['einstein'], ['einstein - 1 match', '@article{einstein,', 'author = {Albert Einstein},'],
         'False'],
        [['einstein', '-i'], [
            'einstein - 2 matches', '@article{einstein,', 'author = {Albert Einstein},',
            'doi = {http://dx.doi.org/10.1002/andp.19053221004},'
        ], 'False'],
        [['einstein', '-i', '-c', '0'], [
            'einstein - 2 matches', '@article{einstein,', 'author = {Albert Einstein},'
        ], 'False'],
        [['einstein', '-i', '-c', '2'], [
            'einstein - 2 matches', '@article{einstein,', 'author = {Albert Einstein},',
            'doi = {http://dx.doi.org/10.1002/andp.19053221004},', 'journal = {Annalen der Physik},'
        ], 'False'],
        [['einstein'], [
            'einstein - 2 matches', '@article{einstein,', 'author = {Albert Einstein},',
            'doi = {http://dx.doi.org/10.1002/andp.19053221004},'
        ], 'True'],
        [['einstein', '-i'], [
            'einstein - 2 matches', '@article{einstein,', 'author = {Albert Einstein},',
            'doi = {http://dx.doi.org/10.1002/andp.19053221004},'
        ], 'True'],
    ])
def test_search(setup, args, expected, config_overwrite):
    """Test search command.

    Args:
        setup: runs pytest fixture.
        args: arguments for the list command call.
        expected: expected result.
        config_overwrite: with what to overwrite the DATABASE/ignore_search_case config option.
    """
    CONFIG.config['DATABASE']['search_ignore_case'] = config_overwrite
    file = StringIO()
    commands.SearchCommand().execute(args, out=file)
    for line, exp in zip_longest(file.getvalue().split('\n'), expected):
        line = line.replace('\x1b', '')
        line = re.sub(r'\[[0-9;]+m', '', line)
        if exp:
            assert exp in line
        if line and not (line.endswith('match') or line.endswith('matches')):
            assert re.match(r'\[[0-9]+\]', line)



"""CoBib auxiliary TextBuffer."""

import curses
import logging
import re
import textwrap

LOGGER = logging.getLogger(__name__)


class TextBuffer:
    """TextBuffer class used as an auxiliary variable to redirect output.

    This buffer class implements a `write` method which allows it to be used as a drop-in source
    for the `file` argument of the `print()` method. Thereby, its output can be gathered in this
    buffer for further usage (such as printing it into a curses pad).
    """

    INDENT = "↪"

    def __init__(self):
        """Initializes the TextBuffer object."""
        self.lines = []
        self.height = 0
        self.width = 0
        self.wrapped = False
        self.ansi_map = None

    def write(self, string):
        """Writes a non-empty string into the buffer.

        Args:
            string (str): the string to append to the buffer.
        """
        if string.strip():
            # only handle non-empty strings
            LOGGER.debug('Appending string to text buffer: %s', string)
            self.lines.append(string)
            self.height = len(self.lines)
            self.width = max(self.width, len(string))

    def replace(self, lines, old_str, new_str):
        """Replaces the old string with the new in the given lines.

        Args:
            lines (int or list): index or indices on which lines to do the replacement.
            old_str (str): old string to be replaced.
            new_str (str): new string to be inserted.
        """
        if isinstance(lines, int):
            lines = [lines]
        for idx in lines:
            self.lines[idx] = re.sub(old_str, new_str, self.lines[idx])

    def flush(self):
        """Compatibility function."""

    def clear(self):
        """Clears the buffer."""
        LOGGER.debug('Clearing text buffer.')
        self.lines = []
        self.height = 0
        self.width = 0
        self.wrapped = False

    def split(self):
        """Split the lines at literal line breaks."""
        copy = self.lines.copy()
        self.lines = []
        self.width = 0
        for line in copy:
            for string in line.split('\n'):
                self.lines.append(string)
                self.width = max(self.width, len(string))
        self.height = len(self.lines)

    def wrap(self, width, label_column=True):
        """Wrap text in buffer to given width.

        Args:
            width (int): maximum text width for each line in the buffer.
            label_column (bool, optional): whether to determine a label column width.
        """
        copy = self.lines.copy()
        self.lines = []
        self.width = 0
        if self.wrapped:
            LOGGER.debug('Unwrapping text buffer.')
            for line in copy:
                # unwrap instead
                if line.startswith(TextBuffer.INDENT):
                    # Note: insert single space when joining lines and strip INDENT symbol
                    self.lines[-1] += ' ' + line[1:].strip()
                else:
                    self.lines.append(line)
                self.width = max(self.width, len(self.lines[-1]))
        else:
            LOGGER.debug('Wrapping text buffer.')
            if label_column:
                # first, determine width of label column
                label_len = 0
                for line in copy:
                    label = line.split('  ')[0]
                    label_len = max(len(label)+1, label_len)
                    if label_len > width:
                        LOGGER.debug('Label column width would exceed actual width. Continuing'
                                     'without a label column.')
                        label_len = 1
                        break
                LOGGER.debug('Label column width determined to be %d', label_len)
            else:
                label_len = 2
                LOGGER.debug('Usig default label column width %d', label_len)
            for line in copy:
                # then wrap lines with subsequent indents matched to first column width
                for string in textwrap.wrap(line, width=width-1,
                                            subsequent_indent=TextBuffer.INDENT + ' ' * label_len):
                    self.lines.append(string)
                self.width = width
        self.height = len(self.lines)
        self.wrapped = not self.wrapped

    # pylint: disable=too-many-arguments
    def view(self, pad, smaxrow, smaxcol,
             pminrow=0, pmincol=0,
             sminrow=1, smincol=0,
             ansi_map=None,
             background=None,
             box=False):
        """View buffer in provided curses pad.

        Args:
            pad (curses.window): a re-sizable curses window (aka a pad).
            smaxrow (int): the available height for the pad.
            smaxcol (int): the available width for the pad.
            pminrow (int): the upper display position in the pad.
            pmincol (int): the left display position in the pad.
            sminrow (int): the upper display position of the pad.
            smincol (int): the left display position of the pad.
            ansi_map (dict): optional, dictionary mapping ANSI codes to curses color pairs.
            background (int): the curses color pair number with which to highlight the window.
            box (bool): whether to print a frame around the window.
        """
        # a regex to detect ANSI color codes
        ansi_regex = re.compile(r'(\x1b\[(\d+)[;]*(\d+)*m)')
        if ansi_map:
            LOGGER.debug('Interpreting ANSI color codes on the fly.')
            self.ansi_map = ansi_map
        # first clear pad
        LOGGER.debug('Clearing curses pad.')
        pad.erase()
        LOGGER.debug('Pad view parameters:')
        LOGGER.debug((pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol))
        pad.refresh(pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol)
        # then resize
        LOGGER.debug('Adjusting pad size.')
        # NOTE The +1 added onto the height accounts for some weird offset in the curses pad.
        pad.resize(self.height+(2 if box else 1), max(self.width, smaxcol+(0 if box else 1)))
        # and populate
        for row, line in enumerate(self.lines):
            start, end, color = -1, -1, -1
            # This list will store the spanned regions for each ANSI color pair.
            # Its entries will be [curses color pair number, start, end].
            color_spans = []
            if self.ansi_map:
                LOGGER.debug('Applying ANSI color map.')
                # The index below is used to keep track of the oldest (i.e. lowest in index) color
                # span which has not been closed yet. This means that we work with the assumption
                # that ANSI color codes always occur in pairs (even though a single \x1b[0m sequence
                # could terminate multiple open spans).
                lowest_incomplete_color_span = 0

                # In order to correctly trim the old line while piecing together the new one we need
                # to keep track of the previously encountered end position.
                prev_end = 0
                new_line = ''

                # iterate over all ANSI color code matches on the current line
                for match in ansi_regex.finditer(line):
                    # add everything preceding the current match to the new line
                    new_line += line[:match.start()-prev_end]
                    # trim the current line from everything in front of the match and itself
                    line = line[match.end()-prev_end:]
                    # store the current end position to use as an offset on the next match
                    prev_end = match.end()

                    # handle ANSI color code
                    if match[0] == '\x1b[0m':
                        # a closing sequence completes the lowest incomplete color span
                        # Note: as mentioned above, we work under the assumption that all ANSI color
                        # codes occur in pairs!
                        color_spans[lowest_incomplete_color_span][2] = len(new_line)
                        lowest_incomplete_color_span += 1
                    else:
                        # else we create a new color span
                        color_spans.append([self.ansi_map[match[0]], len(new_line), None])

                # anything left from the original line needs to be added to the new one
                new_line += line
                # and finally the new line replaces the original one
                line = new_line

            pad.addstr(row+(1 if box else 0), (1 if box else 0), line)
            for color, start, end in sorted(color_spans):
                pad.chgat(row, start, end-start, curses.color_pair(color))
        # apply decorations
        if background is not None:
            # setting background color
            pad.bkgd(' ', curses.color_pair(background + 1))
        if box:
            pad.box()
        LOGGER.debug('Viewing curses pad.')
        pad.refresh(pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol)

    def popup(self, tui, background=None):
        """Displays the buffer in a popup window.

        Args:
            tui (TUI): the TUI instance in which the popup is displayed.
            background (int): the curses color pair number with which to highlight the window.
        """
        LOGGER.debug('Create popup window.')
        popup_win = curses.newpad(self.height+2, tui.width)
        # computing height offset
        height_offset = tui.height - self.height-4
        # wrap buffer prior to viewing
        self.wrap(tui.width, label_column=False)
        # view popup window
        self.view(popup_win, min(height_offset+self.height+2, tui.height-1), tui.width,
                  sminrow=height_offset,
                  background=background, box=True)

        key = 0
        # loop until quit by user
        while key not in (27, ord('q')):  # exit on ESC
            if key != 0:
                # do not print warning on initial run (key == 0)
                tui.prompt_print('To quit the popup window, press "q".')
            key = tui.prompt.getch()

        # close popup window
        popup_win.clear()
        tui.resize_handler(None, None)


class InputBuffer:
    """InputBuffer class used to replace `sys.stdin` and handle user interaction.

    This buffer class implements a `readline` method which allows it to be used as a drop-in
    replacement for `sys.stdin`. It is constructed with a reference to the replacement of
    `sys.stdout` and the TUI instance.
    """

    def __init__(self, buffer, tui):
        """Initializes the InputBuffer object.

        Args:
            buffer (TextBuffer): the current `sys.stdout` replacement.
            tui (TUI): the current TUI instance.
        """
        self.buffer = buffer
        self.tui = tui

    def readline(self):
        """Reads an input from the user mimicking `sys.stdin`."""
        self.buffer.split()
        LOGGER.debug('Create popup window.')
        popup_win = curses.newpad(self.buffer.height+2, self.tui.width)
        # view popup window
        height_offset = self.tui.height - self.buffer.height-4
        # wrap buffer prior to viewing
        self.buffer.wrap(self.tui.width, label_column=False)
        # view buffer
        self.buffer.view(popup_win, height_offset+self.buffer.height+2, self.tui.width,
                         sminrow=height_offset, box=True)

        user_input = self.tui.prompt_handler(None, symbol="> ")

        # close popup window
        popup_win.clear()
        self.buffer.clear()
        self.tui.resize_handler(None, None)

        return user_input
